@@ -34,7 +34,7 @@
 #  Author:   Spencer McIntyre (zeroSteiner)
 
 # Config File Example
-"""
+FILE_CONFIG = """
 [server]
 ip = 0.0.0.0
 port = 8080
@@ -45,27 +45,26 @@ list_directories = True
 """
 
 # The AdvancedHTTPServer systemd service unit file
-"""
 # Quick How To:
 # 1. Copy this file to /etc/systemd/system/pyhttpd.service
-# 2. Edit <USER> and run parameters appropriately in the ExecStart option
+# 2. Edit the run parameters appropriately in the ExecStart option
 # 3. Set configuration settings in /etc/pyhttpd.conf
 # 4. Run "systemctl daemon-reload"
-
+FILE_SYSTEMD_SERVICE_UNIT = """
 [Unit]
 Description=Python Advanced HTTP Server
 After=network.target
 
 [Service]
 Type=simple
-ExecStart=/sbin/runuser -l <USER> -c "/usr/bin/python -m AdvancedHTTPServer -c /etc/pyhttpd.conf"
+ExecStart=/sbin/runuser -l nobody -c "/usr/bin/python -m AdvancedHTTPServer -c /etc/pyhttpd.conf"
 ExecStop=/bin/kill -INT $MAINPID
 
 [Install]
 WantedBy=multi-user.target
 """
 
-__version__ = '0.4.0'
+__version__ = '0.4.2'
 __all__ = [
 	'AdvancedHTTPServer',
 	'AdvancedHTTPServerRegisterPath',
@@ -80,7 +79,6 @@ __all__ = [
 
 import base64
 import binascii
-import cgi
 import datetime
 import hashlib
 import hmac
@@ -108,6 +106,7 @@ import zlib
 
 if sys.version_info[0] < 3:
 	import BaseHTTPServer
+	import cgi as html
 	import Cookie
 	import httplib
 	import SocketServer as socketserver
@@ -118,6 +117,7 @@ if sys.version_info[0] < 3:
 	urllib.parse.unquote = urllib.unquote
 	from ConfigParser import ConfigParser
 else:
+	import html
 	import http.client
 	import http.cookies
 	import http.server
@@ -143,7 +143,7 @@ def _json_object_hook(obj):
 
 SERIALIZER_DRIVERS = {}
 """Dictionary of available drivers for serialization."""
-SERIALIZER_DRIVERS['application/json'] = {'loads': lambda d: json.loads(d, object_hook=_json_object_hook), 'dumps': lambda d: json.dumps(d, default=_json_default)}
+SERIALIZER_DRIVERS['application/json'] = {'loads': lambda d, e: json.loads(d, object_hook=_json_object_hook), 'dumps': lambda d: json.dumps(d, default=_json_default)}
 
 try:
 	import msgpack
@@ -152,17 +152,22 @@ except ImportError:
 else:
 	def _msgpack_default(obj):
 		if isinstance(obj, datetime.datetime):
-			return msgpack.ExtType(10, obj.isoformat())
+			obj = obj.isoformat()
+			if sys.version_info[0] == 3:
+				obj = obj.encode('utf-8')
+			return msgpack.ExtType(10, obj)
 		raise TypeError('Unknown type: ' + repr(obj))
 
 	def _msgpack_ext_hook(code, data):
 		if code == 10:
+			if sys.version_info[0] == 3:
+				data = data.decode('utf-8')
 			if '.' in data:
 				return datetime.datetime.strptime(data, '%Y-%m-%dT%H:%M:%S.%f')
 			else:
 				return datetime.datetime.strptime(data, '%Y-%m-%dT%H:%M:%S')
 		return msfpack.ExtType(code, data)
-	SERIALIZER_DRIVERS['binary/message-pack'] = {'loads': lambda d: msgpack.loads(d, ext_hook=_msgpack_ext_hook), 'dumps': lambda d: msgpack.dumps(d, default=_msgpack_default)}
+	SERIALIZER_DRIVERS['binary/message-pack'] = {'loads': lambda d, e: msgpack.loads(d, encoding=e, ext_hook=_msgpack_ext_hook), 'dumps': lambda d: msgpack.dumps(d, default=_msgpack_default)}
 
 
 if hasattr(logging, 'NullHandler'):
@@ -180,6 +185,27 @@ def random_string(size):
 	"""
 	return ''.join(random.choice(string.ascii_letters + string.digits) for x in range(size))
 
+def resolve_ssl_protocol_version(version=None):
+	"""
+	Look up an SSL protocol version by name. If *version* is not specified, then
+	the strongest protocol available will be returned.
+
+	:param str version: The name of the version to look up.
+	:return: A protocol constant from the :py:mod:`ssl` module.
+	:rtype: int
+	"""
+	if version == None:
+		protocol_preference = ('TLSv1_2', 'TLSv1_1', 'TLSv1', 'SSLv3', 'SSLv23', 'SSLv2')
+		for protocol in protocol_preference:
+			if hasattr(ssl, 'PROTOCOL_' + protocol):
+				return getattr(ssl, 'PROTOCOL_' + protocol)
+		raise RuntimeError('could not find a suitable ssl PROTOCOL_ version constant')
+	elif isinstance(version, str):
+		if not hasattr(ssl, 'PROTOCOL_' + version):
+			raise ValueError('invalid ssl protocol version: ' + version)
+		return getattr(ssl, 'PROTOCOL_' + version)
+	raise TypeError("ssl_version() argument 1 must be str, not {0}".format(type(version).__name__))
+
 def build_serializer_from_content_type(content_type):
 	"""
 	Build a serializer object from a MIME Content-Type string.
@@ -384,8 +410,10 @@ class AdvancedHTTPServerRPCError(Exception):
 	@property
 	def is_remote_exception(self):
 		"""
-		This is true of the represented error resulted from an exception
-		on the remote server.
+		This is true if the represented error resulted from an exception on the
+		remote server.
+
+		:type: bool
 		"""
 		return bool(self.remote_exception != None)
 
@@ -437,7 +465,7 @@ class AdvancedHTTPServerRPCClient(object):
 		:param str serializer_name: The name of the serializer to use.
 		:param str compression: The name of a compression library to use.
 		"""
-		self.serializer = AdvancedHTTPServerSerializer(serializer_name, charset='UTF-8')
+		self.serializer = AdvancedHTTPServerSerializer(serializer_name, charset='UTF-8', compression=compression)
 		self.logger.debug('using serializer: ' + serializer_name)
 
 	def __call__(self, *args, **kwargs):
@@ -477,7 +505,7 @@ class AdvancedHTTPServerRPCClient(object):
 		if self.hmac_key != None:
 			hmac_calculator = hmac.new(self.hmac_key, digestmod=hashlib.sha1)
 			hmac_calculator.update(options)
-			headers['HMAC'] = hmac_calculator.hexdigest()
+			headers['X-RPC-HMAC'] = hmac_calculator.hexdigest()
 
 		if self.username != None and self.password != None:
 			headers['Authorization'] = 'Basic ' + base64.b64encode((self.username + ':' + self.password).encode('UTF-8')).decode('UTF-8')
@@ -492,7 +520,7 @@ class AdvancedHTTPServerRPCClient(object):
 
 		resp_data = resp.read()
 		if self.hmac_key != None:
-			hmac_digest = resp.getheader('hmac')
+			hmac_digest = resp.getheader('X-RPC-HMAC')
 			if not isinstance(hmac_digest, str):
 				raise AdvancedHTTPServerRPCError('hmac validation error', resp.status)
 			hmac_digest = hmac_digest.lower()
@@ -536,7 +564,7 @@ class AdvancedHTTPServerRPCClientCached(AdvancedHTTPServerRPCClient):
 		cursor.execute('SELECT return_value FROM cache WHERE method = ? AND options_hash = ?', (method, options_hash))
 		return_value = cursor.fetchone()
 		if return_value:
-			return_value = self.cache_serializer_loads(return_value[0])
+			return_value = self.cache_serializer_loads(return_value[0], 'UTF-8')
 		else:
 			return_value = self.call(method, *options)
 			cursor.execute('INSERT INTO cache (method, options_hash, return_value) VALUES (?, ?, ?)', (method, options_hash, self.cache_serializer_dumps(return_value)))
@@ -634,9 +662,12 @@ class AdvancedHTTPServerRequestHandler(http.server.BaseHTTPRequestHandler, objec
 
 	def __init__(self, *args, **kwargs):
 		self.handler_map = {}
+		"""The dict object which maps regular expressions of resources to the functions which should handle them."""
 		self.rpc_handler_map = {}
+		"""The dict object which maps regular expressions of RPC functions to their handlers."""
 		self.server = args[2]
 		self.headers_active = False
+		"""Whether or not the request is in the sending headers phase."""
 		for map_name in (None, self.__class__.__name__):
 			handler_map = GLOBAL_HANDLER_MAP.get(map_name, {})
 			for path, function_info in handler_map.items():
@@ -648,6 +679,11 @@ class AdvancedHTTPServerRequestHandler(http.server.BaseHTTPRequestHandler, objec
 		self.install_handlers()
 
 		self.basic_auth_user = None
+		"""The name of the user if the current request is using basic authentication."""
+		self.query_data = None
+		"""The parameter data that has been passed to the server parsed as a dict."""
+		self.raw_query_data = None
+		"""The raw data that was parsed into the :py:attr:`.query_data` attribute."""
 		super(AdvancedHTTPServerRequestHandler, self).__init__(*args, **kwargs)
 
 	def version_string(self):
@@ -669,6 +705,7 @@ class AdvancedHTTPServerRequestHandler(http.server.BaseHTTPRequestHandler, objec
 		:param str file_path: The path to the file to serve, this does not need to be in the web root.
 		:param bool attachment: Whether to serve the file as a download by setting the Content-Disposition header.
 		"""
+		del query
 		file_path = os.path.abspath(file_path)
 		try:
 			file_obj = open(file_path, 'rb')
@@ -695,6 +732,7 @@ class AdvancedHTTPServerRequestHandler(http.server.BaseHTTPRequestHandler, objec
 
 		:param str dir_path: The path of the directory to list the contents of.
 		"""
+		del query
 		try:
 			dir_contents = os.listdir(dir_path)
 		except os.error:
@@ -703,7 +741,7 @@ class AdvancedHTTPServerRequestHandler(http.server.BaseHTTPRequestHandler, objec
 		if os.path.normpath(dir_path) != self.server.serve_files_root:
 			dir_contents.append('..')
 		dir_contents.sort(key=lambda a: a.lower())
-		displaypath = cgi.escape(urllib.parse.unquote(self.path))
+		displaypath = html.escape(urllib.parse.unquote(self.path), quote=True)
 
 		f = io.BytesIO()
 		encoding = sys.getfilesystemencoding()
@@ -721,7 +759,7 @@ class AdvancedHTTPServerRequestHandler(http.server.BaseHTTPRequestHandler, objec
 			if os.path.islink(fullname):
 				displayname = name + "@"
 				# Note: a link to a directory displays with @ and links with /
-			f.write(('<li><a href="' + urllib.parse.quote(linkname) + '">' + cgi.escape(displayname) + '</a>\n').encode(encoding))
+			f.write(('<li><a href="' + urllib.parse.quote(linkname) + '">' + html.escape(displayname, quote=True) + '</a>\n').encode(encoding))
 		f.write(b'</ul>\n<hr>\n</body>\n</html>\n')
 		length = f.tell()
 		f.seek(0)
@@ -909,18 +947,22 @@ class AdvancedHTTPServerRequestHandler(http.server.BaseHTTPRequestHandler, objec
 
 	def stock_handler_respond_unauthorized(self, query):
 		"""This method provides a handler suitable to be used in the handler_map."""
+		del query
 		self.respond_unauthorized()
 		return
 
 	def stock_handler_respond_not_found(self, query):
 		"""This method provides a handler suitable to be used in the handler_map."""
+		del query
 		self.respond_not_found()
 		return
 
 	def check_authorization(self):
 		"""
 		Check for the presence of a basic auth Authorization header and
-		if the credentials contained within are valid.
+		if the credentials contained within in are valid.
+		:return: Whether or not the credentials are valid.
+		:rtype: bool
 		"""
 		try:
 			if self.server.basic_auth == None:
@@ -996,9 +1038,7 @@ class AdvancedHTTPServerRequestHandler(http.server.BaseHTTPRequestHandler, objec
 
 		self.dispatch_handler(self.query_data)
 		return
-
-	def do_HEAD(self):
-		self.do_GET()
+	do_HEAD = do_GET
 
 	def do_POST(self):
 		if not self.check_authorization():
@@ -1006,11 +1046,13 @@ class AdvancedHTTPServerRequestHandler(http.server.BaseHTTPRequestHandler, objec
 			return
 		content_length = int(self.headers.get('content-length', 0))
 		data = self.rfile.read(content_length)
-		self.query_data_raw = data
+		self.raw_query_data = data
 		content_type = self.headers.get('content-type', '')
 		content_type = content_type.split(';', 1)[0]
 		self.query_data = {}
 		try:
+			if not isinstance(data, str):
+				data = data.decode(self.get_content_type_charset())
 			if content_type.startswith('application/json'):
 				data = json.loads(data)
 				if isinstance(data, dict):
@@ -1024,7 +1066,7 @@ class AdvancedHTTPServerRequestHandler(http.server.BaseHTTPRequestHandler, objec
 		return
 
 	def do_OPTIONS(self):
-		available_methods = map(lambda x: x[3:], filter(lambda x: x.startswith('do_'), dir(self)))
+		available_methods = list(map(lambda x: x[3:], filter(lambda x: x.startswith('do_'), dir(self))))
 		if 'RPC' in available_methods and len(self.rpc_handler_map) == 0:
 			available_methods.remove('RPC')
 		self.send_response(200)
@@ -1054,7 +1096,7 @@ class AdvancedHTTPServerRequestHandler(http.server.BaseHTTPRequestHandler, objec
 			return
 
 		if self.server.rpc_hmac_key != None:
-			hmac_digest = self.headers.get('hmac')
+			hmac_digest = self.headers.get('X-RPC-HMAC')
 			if not isinstance(hmac_digest, str):
 				self.respond_unauthorized(request_authentication=True)
 				return
@@ -1112,7 +1154,7 @@ class AdvancedHTTPServerRequestHandler(http.server.BaseHTTPRequestHandler, objec
 		if self.server.rpc_hmac_key != None:
 			hmac_calculator = hmac.new(self.server.rpc_hmac_key, digestmod=hashlib.sha1)
 			hmac_calculator.update(response)
-			self.send_header('HMAC', hmac_calculator.hexdigest())
+			self.send_header('X-RPC-HMAC', hmac_calculator.hexdigest())
 		self.end_headers()
 
 		self.wfile.write(response)
@@ -1124,6 +1166,32 @@ class AdvancedHTTPServerRequestHandler(http.server.BaseHTTPRequestHandler, objec
 	def log_message(self, format, *args):
 		self.server.logger.info(self.address_string() + ' ' + format % args)
 
+	def get_query(self, name, default=None):
+		"""
+		Get a value from the query data that was sent to the server.
+
+		:param str name: The name of the query value to retrieve.
+		:return: The value if it exists, otherwise *default* will be returned.
+		:rtype: str
+		"""
+		return self.query_data.get(name, [default])[0]
+
+	def get_content_type_charset(self, default='UTF-8'):
+		"""
+		Inspect the Content-Type header to retrieve the charset that the client
+		has specified.
+
+		:param str default: The default charset to return if none exists.
+		:return: The charset of the request.
+		:rtype: str
+		"""
+		encoding = default
+		header = self.headers.get('Content-Type', '')
+		idx = header.find('charset=')
+		if idx > 0:
+			encoding = (header[idx + 8:].split(' ', 1)[0] or encoding)
+		return encoding
+
 class AdvancedHTTPServerSerializer(object):
 	"""
 	This class represents a serilizer object for use with the RPC system.
@@ -1172,7 +1240,7 @@ class AdvancedHTTPServerSerializer(object):
 			data = zlib.decompress(data)
 		if sys.version_info[0] == 3 and self.name.startswith('application/'):
 			data = data.decode(self._charset)
-		data = SERIALIZER_DRIVERS[self.name]['loads'](data)
+		data = SERIALIZER_DRIVERS[self.name]['loads'](data, (self._charset if sys.version_info[0] == 3 else None))
 		if isinstance(data, list):
 			data = tuple(data)
 		return data
@@ -1185,7 +1253,7 @@ class AdvancedHTTPServer(object):
 	the port is guessed based on if the server is run as root or not and
 	SSL is used.
 	"""
-	def __init__(self, RequestHandler, address=None, use_threads=True, ssl_certfile=None, ssl_keyfile=None):
+	def __init__(self, RequestHandler, address=None, use_threads=True, ssl_certfile=None, ssl_keyfile=None, ssl_version=None):
 		"""
 		:param RequestHandler: The request handler class to use.
 		:type RequestHandler: :py:class:`.AdvancedHTTPServerRequestHandler`
@@ -1193,6 +1261,7 @@ class AdvancedHTTPServer(object):
 		:param bool use_threads: Whether to enable the use of a threaded handler.
 		:param str ssl_certfile: An SSL certificate file to use, setting this enables SSL.
 		:param str ssl_keyfile: An SSL certificate file to use.
+		:param ssl_version: The SSL protocol version to use.
 		"""
 		self.use_ssl = bool(ssl_certfile)
 		if address == None:
@@ -1220,7 +1289,9 @@ class AdvancedHTTPServer(object):
 		self.logger.info('listening on ' + address[0] + ':' + str(address[1]))
 
 		if self.use_ssl:
-			self.http_server.socket = ssl.wrap_socket(self.http_server.socket, keyfile=ssl_keyfile, certfile=ssl_certfile, server_side=True)
+			if ssl_version == None or isinstance(ssl_version, str):
+				ssl_version = resolve_ssl_protocol_version(ssl_version)
+			self.http_server.socket = ssl.wrap_socket(self.http_server.socket, keyfile=ssl_keyfile, certfile=ssl_certfile, server_side=True, ssl_version=ssl_version)
 			self.http_server.using_ssl = True
 			self.logger.info(address[0] + ':' + str(address[1]) + ' - ssl has been enabled')
 
@@ -1375,7 +1446,8 @@ class AdvancedHTTPServer(object):
 		as a hash by specifying the hash type in the *pwtype* argument.
 
 		:param str username: The username of the credentials to be added.
-		:param bytes, str password: The password data of the credentials to be added.
+		:param password: The password data of the credentials to be added.
+		:type password: bytes, str
 		:param str pwtype: The type of the *password* data, (plain, md5, sha1, etc.).
 		"""
 		if not isinstance(password, (bytes, str)):
@@ -1446,6 +1518,7 @@ class AdvancedHTTPServerTestCase(unittest.TestCase):
 		self.http_connection = http.client.HTTPConnection(self.server_address[0], self.server_address[1])
 
 	def _test_resource_handler(self, handler, query):
+		del query
 		handler.send_response(200)
 		handler.end_headers()
 		message = b'Hello World!\r\n\r\n'
